extern crate native_tls;
extern crate rand;

use native_tls::{Identity, TlsAcceptor, TlsStream};
use rand::seq::SliceRandom;
use std::env;
use std::fs::File;
use std::fs::OpenOptions;
use std::io::{Read, Write};
use std::net::{TcpListener, TcpStream};
use std::str;
use std::sync::Arc;
use std::thread;
use url::Url;
use percent_encoding::percent_decode;

extern crate chrono;
use chrono::{Datelike, Timelike, Utc};

fn main() {
    let args: Vec<String> = env::args().collect();
    let pkcs12 = make_pkcs12(&args[1], &args[2]);

    let acceptor = TlsAcceptor::new(pkcs12).unwrap();
    let acceptor = Arc::new(acceptor);

    let listener = TcpListener::bind("0.0.0.0:1965").unwrap();

    println!("\n=> Started pollux on port 1965. Awaiting connections...\n");

    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                println!("New connection: {}", stream.peer_addr().unwrap());
                let acceptor = acceptor.clone();
                thread::spawn(move || {
                    let stream = acceptor.accept(stream).unwrap();
                    handle_request(stream);
                });
            }
            Err(e) => {
                println!("Error: {}", e);
            }
        }
    }

    drop(listener);
}

fn handle_request(mut stream: TlsStream<TcpStream>) {
    let mut buffer = [0; 512];
    stream.read(&mut buffer).unwrap();
    let request = String::from_utf8_lossy(&buffer[..]).to_owned();
    println!("Request: {}", request);

    route(stream, request.to_string());
}

fn respond_input(mut stream: TlsStream<TcpStream>, input: String) {
    stream
        .write_all(format!("10\t{}\r\n", input).as_bytes())
        .unwrap();
}

fn respond_success(mut stream: TlsStream<TcpStream>, content: String, query: Option<&str>) {
    // a header
    stream.write_all(b"20\ttext/gemini\r\n").unwrap();

    match query {
        Some(query) => {
            append_guestbook(query);
            let guestbook = read_guestbook().unwrap();
            stream
                .write_all(guestbook.as_bytes())
                .expect("Unable to serve file");
        }
        None => {
            stream
                .write_all(content.as_bytes())
                .expect("Unable to serve file");
        }
    }
    // a footer :)
    stream
        .write_all(b"\n\n---\nServed by Pollux Gemini Server.")
        .unwrap();
}

fn respond_binary(mut stream: TlsStream<TcpStream>, path: &str, mime_type: &str) {
    let mut f = File::open(path).unwrap();
    let mut buffer = Vec::new();
    f.read_to_end(&mut buffer).unwrap();

    let header = format!("20\t{}\r\n", mime_type);
    stream.write_all(header.as_bytes()).unwrap();
    stream.write_all(&buffer).unwrap();
}

fn respond_redirect(mut stream: TlsStream<TcpStream>) {
    let url = random_url();
    stream
        .write_all(format!("31\t{}\r\n", url).as_bytes())
        .unwrap();
}

fn respond_not_found(mut stream: TlsStream<TcpStream>) {
    stream
        .write_all("51\tNot found\r\n".as_bytes())
        .unwrap();
}

fn append_guestbook(message: &str) {
    let mut file = OpenOptions::new()
        .create(true)
        .append(true)
        .write(true)
        .open("/var/gemini/guestbook")
        .expect("file not found");

    let now = Utc::now();
    let date = format!("{:02}/{:02}/{:04} at {:02}:{:02}", now.day(), now.month(), now.year(), now.hour(), now.minute());
    let text = str::replace(message, "query=", "");
    let utf8_text = percent_decode(text.as_bytes()).decode_utf8().unwrap();
    let entry = format!("\n- {}\n{:?}\n", date, utf8_text);
    file.write_all(entry.as_bytes())
        .expect("Unable to write file");
}

fn read_guestbook() -> Option<String> {
    let mut file = File::open("/var/gemini/guestbook").expect("file not found");
    let mut content = String::new();
    file.read_to_string(&mut content)
        .expect("Unable to read file");
    if content.len() > 0 {
        Some(content)
    } else {
        None
    }
}

fn random_url() -> &'static str {
    let urls = vec![
        "gemini://gemini.conman.org",
        "gemini://zaibatsu.circumlunar.space",
        "gemini://carcosa.net",
        "gemini://heavysquare.com",
        "gemini://mozz.us",
        "gemini://dgold.eu",
        "gemini://tilde.black",
        "gemini://tilde.pink",
        "gemini://vger.cloud",
        "gemini://earthlight.xyz",
        "gemini://yam655.com",
    ];
    let url = urls.choose(&mut rand::thread_rng()).unwrap();
    url.to_owned()
}

fn make_pkcs12(pfx_file_path: &str, pfx_file_password: &str) -> native_tls::Identity {
    let mut file = File::open(pfx_file_path).expect("pfx file not found!");
    let mut pkcs12 = vec![];
    file.read_to_end(&mut pkcs12).unwrap();

    Identity::from_pkcs12(&pkcs12, pfx_file_password).unwrap()
}

fn route(stream: TlsStream<TcpStream>, request: String) {
    let url = Url::parse(&request).unwrap();
    let query = url.query();

    let route = match url.path() {
        "/" => (20, String::from("/index.gemini"), None),
        "/roulette" => (31, String::new(), None),
        "/guestbook" => (20, String::from("/guestbook"), None),
        "/guestbook/new" => {
            if query.is_some() {
                (20, String::from("/guestbook"), query)
            } else {
                (10, String::from("Leave a message."), None)
            }
        },
        "/project_gemini.jpg" => (22, String::from("/project_gemini.jpg"), None),
        path => (20, String::from(path), None),
    };

    match route {
        (10, message, _) => respond_input(stream, String::from(message)),
        (20, path, query) => {
            let full_path = if path.starts_with("/var/gemini") {
                path
            } else {
                format!("/var/gemini{}", path)
            };

            let open_file = OpenOptions::new()
                .read(true)
                .open(&full_path);

            match open_file {
                Ok(mut file) => {
                    let mut content = String::new();
                    file.read_to_string(&mut content).expect("Can't read content.");

                    respond_success(stream, content, query)
                }
                Err(_) => {
                    respond_not_found(stream)
                }
            }
        }
        (22, path, _query) => {
            let full_path = if path.starts_with("/var/gemini") {
                path
            } else {
                format!("/var/gemini{}", path)
            };

            respond_binary(stream, &full_path, "image/jpeg")
        }
        (31, _, _) => respond_redirect(stream),
        _ => (),
    }
}
